/* $XConsortium: ThreeList.c,v 1.34 91/09/27 18:35:07 converse Exp $ */

/*
 * Copyright 1989 Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of M.I.T. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  M.I.T. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * M.I.T. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL M.I.T.
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

/*
 * ThreeList.c - ThreeList widget
 *
 * This is the ThreeList widget, it is useful to display a list, without the
 * overhead of having a widget for each item in the list.  It allows 
 * the user to select an item in a list and notifies the application through
 * a callback function.
 *
 *	Created: 	8/13/88
 *	By:		Chris D. Peterson
 *                      MIT X Consortium
 */

#include <stdio.h>
#include <ctype.h>

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include <X11/Xmu/Drawing.h>

#include <X11/Xaw/XawInit.h>
#include <X11/Xaw/Viewport.h>
#include "ThreeListP.h"

#include "libsx.h"
#include "libsx_private.h"


extern WindowState *lsx_curwin;   /* global handle to current window */

typedef void (*ThreeListCB)();

/*
 * this structure maintains some internal state information about each
 * scrolled list widget.
 */
typedef struct ListInfo
{
  Widget w;
  ThreeListCB func;
/*
  void (*func)(Widget w, char *str, int index, unsigned int event, void *data);
*/
  void *data;
  struct ListInfo *next;
}ListInfo;

static ListInfo *scroll_lists = NULL;

/* 
 * Default Translation table.
 */

static char defaultTranslations[] =  
  "<Btn1Down>:   Set()\n\
   <Btn1Up>:     Notify()\n\
   <Btn2Down>:   Set()\n\
   <Btn2Up>:     Notify()\n\
   <Btn3Down>:   Set()\n\
   <Btn3Up>:     Notify()";

/****************************************************************
 *
 * Full class record constant
 *
 ****************************************************************/

/* Private Data */

#define offset(field) XtOffset(ThreeListWidget, field)

static XtResource resources[] = {
    {XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
	offset(threeList.foreground), XtRString, XtDefaultForeground},
    {XtNcursor, XtCCursor, XtRCursor, sizeof(Cursor),
       offset(simple.cursor), XtRString, "left_ptr"},
    {XtNfont,  XtCFont, XtRFontStruct, sizeof(XFontStruct *),
	offset(threeList.font),XtRString, XtDefaultFont},
    {XtNlist, XtCList, XtRPointer, sizeof(char **),
       offset(threeList.list), XtRString, NULL},
    {XtNdefaultColumns, XtCColumns, XtRInt,  sizeof(int),
	offset(threeList.default_cols), XtRImmediate, (XtPointer)2},
    {XtNlongest, XtCLongest, XtRInt,  sizeof(int),
	offset(threeList.longest), XtRImmediate, (XtPointer)0},
    {XtNnumberStrings, XtCNumberStrings, XtRInt,  sizeof(int),
	offset(threeList.nitems), XtRImmediate, (XtPointer)0},
    {XtNpasteBuffer, XtCBoolean, XtRBoolean,  sizeof(Boolean),
	offset(threeList.paste), XtRImmediate, (XtPointer) False},
    {XtNforceColumns, XtCColumns, XtRBoolean,  sizeof(Boolean),
	offset(threeList.force_cols), XtRImmediate, (XtPointer) False},
    {XtNverticalList, XtCBoolean, XtRBoolean,  sizeof(Boolean),
	offset(threeList.vertical_cols), XtRImmediate, (XtPointer) False},
    {XtNinternalWidth, XtCWidth, XtRDimension,  sizeof(Dimension),
	offset(threeList.internal_width), XtRImmediate, (XtPointer)4},
    {XtNinternalHeight, XtCHeight, XtRDimension, sizeof(Dimension),
	offset(threeList.internal_height), XtRImmediate, (XtPointer)2},
    {XtNcolumnSpacing, XtCSpacing, XtRDimension,  sizeof(Dimension),
	offset(threeList.column_space), XtRImmediate, (XtPointer)6},
    {XtNrowSpacing, XtCSpacing, XtRDimension,  sizeof(Dimension),
	offset(threeList.row_space), XtRImmediate, (XtPointer)2},
    {XtNcallback, XtCCallback, XtRCallback, sizeof(XtPointer),
        offset(threeList.callback), XtRCallback, NULL},
};

static void Initialize();
static void ChangeSize();
static void Resize();
static void Redisplay();
static void Destroy();
static Boolean Layout();
static XtGeometryResult PreferredGeom();
static Boolean SetValues();
static void Notify(), Set(), Unset();

static XtActionsRec actions[] = {
      {"Notify",         Notify},
      {"Set",            Set},
      {"Unset",          Unset},
};

ThreeListClassRec threeListClassRec = {
  {
/* core_class fields */	
#define superclass		(&simpleClassRec)
    /* superclass	  	*/	(WidgetClass) superclass,
    /* class_name	  	*/	"ThreeList",
    /* widget_size	  	*/	sizeof(ThreeListRec),
    /* class_initialize   	*/	XawInitializeWidgetSet,
    /* class_part_initialize	*/	NULL,
    /* class_inited       	*/	FALSE,
    /* initialize	  	*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize		  	*/	XtInheritRealize,
    /* actions		  	*/	actions,
    /* num_actions	  	*/	XtNumber(actions),
    /* resources	  	*/	resources,
    /* num_resources	  	*/	XtNumber(resources),
    /* xrm_class	  	*/	NULLQUARK,
    /* compress_motion	  	*/	TRUE,
    /* compress_exposure  	*/	FALSE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest	  	*/	FALSE,
    /* destroy		  	*/	Destroy,
    /* resize		  	*/	Resize,
    /* expose		  	*/	Redisplay,
    /* set_values	  	*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus	 	*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private   	*/	NULL,
    /* tm_table		   	*/	defaultTranslations,
   /* query_geometry		*/      PreferredGeom,
  },
/* Simple class fields initialization */
  {
    /* change_sensitive		*/	XtInheritChangeSensitive
  }
};

WidgetClass threeListWidgetClass = (WidgetClass)&threeListClassRec;

/****************************************************************
 *
 * Private Procedures
 *
 ****************************************************************/

static void GetGCs(w)
Widget w;
{
    XGCValues	values;
    ThreeListWidget lw = (ThreeListWidget) w;    

    values.foreground	= lw->threeList.foreground;
    values.font		= lw->threeList.font->fid;
    lw->threeList.normgc = XtGetGC(w, (unsigned) GCForeground | GCFont,
				 &values);

    values.foreground	= lw->core.background_pixel;
    lw->threeList.revgc = XtGetGC(w, (unsigned) GCForeground | GCFont,
				 &values);

    values.tile       = XmuCreateStippledPixmap(XtScreen(w), 
						lw->threeList.foreground,
						lw->core.background_pixel,
						lw->core.depth);
    values.fill_style = FillTiled;

    lw->threeList.graygc = XtGetGC(w, (unsigned) GCFont | GCTile | GCFillStyle,
			      &values);
}

/*	Function Name: ResetThreeList
 *	Description: Resets the new list when important things change.
 *	Arguments: w - the widget.
 *                 changex, changey - allow the height or width to change?
 *	Returns: none.
 */

static void
ResetThreeList(w, changex, changey)
Widget w;
Boolean changex, changey;
{
    ThreeListWidget lw = (ThreeListWidget) w;
    Dimension width = w->core.width;
    Dimension height = w->core.height;
    register int i, len;

/*
 * If list is NULL then the list will just be the name of the widget.
 */

    if (lw->threeList.list == NULL) {
      lw->threeList.list = &(lw->core.name);
      lw->threeList.nitems = 1;
    }

    if (lw->threeList.nitems == 0)	    /* Get number of items. */
        for ( ; lw->threeList.list[lw->threeList.nitems] != NULL ; lw->threeList.nitems++);

    if (lw->threeList.longest == 0) /* Get column width. */
        for ( i = 0 ; i < lw->threeList.nitems; i++) {
	    len = XTextWidth(lw->threeList.font, lw->threeList.list[i],
			     strlen(lw->threeList.list[i]));
	    if (len > lw->threeList.longest)
	        lw->threeList.longest = len;
	}

    lw->threeList.col_width = lw->threeList.longest + lw->threeList.column_space;

    if (Layout(w, changex, changey, &width, &height))
      ChangeSize(w, width, height);
}

/*	Function Name: ChangeSize.
 *	Description: Laysout the widget.
 *	Arguments: w - the widget to try change the size of.
 *	Returns: none.
 */

static void
ChangeSize(w, width, height)
Widget w;
Dimension width, height;
{
    XtWidgetGeometry request, reply;

    request.request_mode = CWWidth | CWHeight;
    request.width = width;
    request.height = height;
    
    switch ( XtMakeGeometryRequest(w, &request, &reply) ) {
    case XtGeometryYes:
    case XtGeometryNo:
        break;
    case XtGeometryAlmost:
	Layout(w, (request.height != reply.height),
	          (request.width != reply.width),
	       &(reply.width), &(reply.height));
	request = reply;
	switch (XtMakeGeometryRequest(w, &request, &reply) ) {
	case XtGeometryYes:
	case XtGeometryNo:
	    break;
	case XtGeometryAlmost:
	    request = reply;
	    if (Layout(w, FALSE, FALSE,
		       &(request.width), &(request.height))) {
	      char buf[BUFSIZ];
	      sprintf(buf, "ThreeList Widget: %s %s",
		      "Size Changed when it shouldn't have",
		      "when computing layout");
	      XtAppWarning(XtWidgetToApplicationContext(w), buf);
	    }
	    request.request_mode = CWWidth | CWHeight;
	    XtMakeGeometryRequest(w, &request, &reply);
	    break;
	default:
	  XtAppWarning(XtWidgetToApplicationContext(w),
		       "ThreeList Widget: Unknown geometry return.");
	  break;
	}
	break;
    default:
	XtAppWarning(XtWidgetToApplicationContext(w),
		     "ThreeList Widget: Unknown geometry return.");
	break;
    }
}

/*	Function Name: Initialize
 *	Description: Function that initilizes the widget instance.
 *	Arguments: junk - NOT USED.
 *                 new  - the new widget.
 *	Returns: none
 */

/* ARGSUSED */
static void 
Initialize(junk, new)
Widget junk, new;
{
    ThreeListWidget lw = (ThreeListWidget) new;

/* 
 * Initialize all private resources.
 */

    GetGCs(new);

    /* Set row height. */
    lw->threeList.row_height = lw->threeList.font->max_bounds.ascent
			+ lw->threeList.font->max_bounds.descent
			+ lw->threeList.row_space;

    ResetThreeList(new, (new->core.width == 0), (new->core.height == 0));

    lw->threeList.highlight = lw->threeList.is_highlighted = NO_HIGHLIGHT;

} /* Initialize */

/*	Function Name: CvtToItem
 *	Description: Converts Xcoord to item number of item containing that
 *                   point.
 *	Arguments: w - the threeList widget.
 *                 xloc, yloc - x location, and y location.
 *	Returns: the item number.
 */

static int
CvtToItem(w, xloc, yloc, item)
Widget w;
int xloc, yloc;
int *item;
{
    int one, another;
    ThreeListWidget lw = (ThreeListWidget) w;
    int ret_val = OKAY;

    if (lw->threeList.vertical_cols) {
        one = lw->threeList.nrows * ((xloc - (int) lw->threeList.internal_width)
	    / lw->threeList.col_width);
        another = (yloc - (int) lw->threeList.internal_height) 
	        / lw->threeList.row_height;
	 /* If out of range, return minimum possible value. */
	if (another >= lw->threeList.nrows) {
	    another = lw->threeList.nrows - 1;
	    ret_val = OUT_OF_RANGE;
	}
    }
    else {
        one = (lw->threeList.ncols * ((yloc - (int) lw->threeList.internal_height) 
              / lw->threeList.row_height)) ;
	/* If in right margin handle things right. */
        another = (xloc - (int) lw->threeList.internal_width) / lw->threeList.col_width;
	if (another >= lw->threeList.ncols) {
	    another = lw->threeList.ncols - 1; 
	    ret_val = OUT_OF_RANGE;
	}
    }  
    if ((xloc < 0) || (yloc < 0))
        ret_val = OUT_OF_RANGE;
    if (one < 0) one = 0;
    if (another < 0) another = 0;
    *item = one + another;
    if (*item >= lw->threeList.nitems) return(OUT_OF_RANGE);
    return(ret_val);
}

/*	Function Name: FindCornerItems.
 *	Description: Find the corners of the rectangle in item space.
 *	Arguments: w - the threeList widget.
 *                 event - the event structure that has the rectangle it it.
 *                 ul_ret, lr_ret - the corners ** RETURNED **.
 *	Returns: none.
 */

static void
FindCornerItems(w, event, ul_ret, lr_ret)
Widget w;
XEvent * event;
int *ul_ret, *lr_ret;
{
    int xloc, yloc;

    xloc = event->xexpose.x;
    yloc = event->xexpose.y;
    CvtToItem(w, xloc, yloc, ul_ret);
    xloc += event->xexpose.width;
    yloc += event->xexpose.height;
    CvtToItem(w, xloc, yloc, lr_ret);
}

/*	Function Name: ItemInRectangle
 *	Description: returns TRUE if the item passed is in the given rectangle.
 *	Arguments: w - the threeList widget.
 *                 ul, lr - corners of the rectangle in item space.
 *                 item - item to check.
 *	Returns: TRUE if the item passed is in the given rectangle.
 */

static Boolean
ItemInRectangle(w, ul, lr, item)
Widget w;
int ul, lr, item;
{
    ThreeListWidget lw = (ThreeListWidget) w;
    register int mod_item;
    int things;
    
    if (item < ul || item > lr) 
        return(FALSE);
    if (lw->threeList.vertical_cols)
        things = lw->threeList.nrows;
    else
        things = lw->threeList.ncols;

    mod_item = item % things;
    if ( (mod_item >= ul % things) && (mod_item <= lr % things ) )
        return(TRUE);
    return(FALSE);
}

/*	Function Name: HighlightBackground
 *	Description: paints the color of the background for the given item.
 *	Arguments: w - the widget.
 *                 x, y - ul corner of the area item occupies.
 *                 item - the item we are dealing with.
 *                 gc - the gc that is used to paint this rectangle
 *	Returns: 
 */

static void
HighlightBackground(w, x, y, item, gc)
Widget w;
int x, y, item;
GC gc;
{
    ThreeListWidget lw = (ThreeListWidget) w;
    int hl_x, hl_y, width, height;

    hl_x = x - lw->threeList.column_space/2;
    width = XTextWidth(lw->threeList.font, lw->threeList.list[item],
			 strlen(lw->threeList.list[item])) + lw->threeList.column_space;
    hl_y = y - lw->threeList.row_space/2;
    height = lw->threeList.row_height + lw->threeList.row_space;

    XFillRectangle(XtDisplay(w), XtWindow(w), gc, hl_x, hl_y, width, height);
}

/*	Function Name: PaintItemName
 *	Description: paints the name of the item in the appropriate location.
 *	Arguments: w - the threeList widget.
 *                 item - the item to draw.
 *	Returns: none.
 *
 *      NOTE: no action taken on an unrealized widget.
 */

static void
PaintItemName(w, item)
Widget w;
int item;
{
    char * str;
    GC gc;
    int x, y, str_y;
    ThreeListWidget lw = (ThreeListWidget) w;

    if (!XtIsRealized(w)) return; /* Just in case... */
   
    if (lw->threeList.vertical_cols) {
	x = lw->threeList.col_width * (item / lw->threeList.nrows)
	  + lw->threeList.internal_width;
        y = lw->threeList.row_height * (item % lw->threeList.nrows)
	  + lw->threeList.internal_height;
    }
    else {
        x = lw->threeList.col_width * (item % lw->threeList.ncols)
	  + lw->threeList.internal_width;
        y = lw->threeList.row_height * (item / lw->threeList.ncols)
	  + lw->threeList.internal_height;
    }

    str_y = y + lw->threeList.font->max_bounds.ascent;

    if (item == lw->threeList.is_highlighted) {
        if (item == lw->threeList.highlight) {
            gc = lw->threeList.revgc;
	    HighlightBackground(w, x, y, item, lw->threeList.normgc);
	}
        else {
	    if (XtIsSensitive(w)) 
	        gc = lw->threeList.normgc;
	    else
	        gc = lw->threeList.graygc;
	    HighlightBackground(w, x, y, item, lw->threeList.revgc);
	    lw->threeList.is_highlighted = NO_HIGHLIGHT;
        }
    }
    else {
        if (item == lw->threeList.highlight) {
            gc = lw->threeList.revgc;
	    HighlightBackground(w, x, y, item, lw->threeList.normgc);
	    lw->threeList.is_highlighted = item;
	}
	else {
	    if (XtIsSensitive(w)) 
	        gc = lw->threeList.normgc;
	    else
	        gc = lw->threeList.graygc;
	}
    }

    str =  lw->threeList.list[item];	/* draw it */
    XDrawString(XtDisplay(w), XtWindow(w), gc, x, str_y, str, strlen(str));
}
    
/*	Function Name: Redisplay
 *	Description: Repaints the widget window on expose events.
 *	Arguments: w - the threeList widget.
 *                 event - the expose event for this repaint.
 *                 junk - NOT USED.
 *	Returns: 
 */

/* ARGSUSED */
static void 
Redisplay(w, event, junk)
Widget w;
XEvent *event;
Region junk;
{
    int item;			/* an item to work with. */
    int ul_item, lr_item;       /* corners of items we need to paint. */
    ThreeListWidget lw = (ThreeListWidget) w;

    if (event == NULL) {	/* repaint all. */
        ul_item = 0;
	lr_item = lw->threeList.nrows * lw->threeList.ncols - 1;
	XClearWindow(XtDisplay(w), XtWindow(w));
    }
    else
        FindCornerItems(w, event, &ul_item, &lr_item);
    
    for (item = ul_item; (item <= lr_item && item < lw->threeList.nitems) ; item++)
      if (ItemInRectangle(w, ul_item, lr_item, item))
	PaintItemName(w, item);
}

/*	Function Name: PreferredGeom
 *	Description: This tells the parent what size we would like to be
 *                   given certain constraints.
 *	Arguments: w - the widget.
 *                 intended - what the parent intends to do with us.
 *                 requested - what we want to happen.
 *	Returns: none.
 */

static XtGeometryResult 
PreferredGeom(w, intended, requested)
Widget w;
XtWidgetGeometry *intended, *requested;
{
    Dimension new_width, new_height;
    Boolean change, width_req, height_req;
    
    width_req = intended->request_mode & CWWidth;
    height_req = intended->request_mode & CWHeight;

    if (width_req)
      new_width = intended->width;
    else
      new_width = w->core.width;

    if (height_req)
      new_height = intended->height;
    else
      new_height = w->core.height;

    requested->request_mode = 0;
    
/*
 * We only care about our height and width.
 */

    if ( !width_req && !height_req)
      return(XtGeometryYes);
    
    change = Layout(w, !width_req, !height_req, &new_width, &new_height);

    requested->request_mode |= CWWidth;
    requested->width = new_width;
    requested->request_mode |= CWHeight;
    requested->height = new_height;

    if (change)
        return(XtGeometryAlmost);
    return(XtGeometryYes);
}

/*	Function Name: Resize
 *	Description: resizes the widget, by changing the number of rows and
 *                   columns.
 *	Arguments: w - the widget.
 *	Returns: none.
 */

static void
Resize(w)
Widget w;
{
  Dimension width, height;

  width = w->core.width;
  height = w->core.height;

  if (Layout(w, FALSE, FALSE, &width, &height))
    XtAppWarning(XtWidgetToApplicationContext(w),
	    "ThreeList Widget: Size changed when it shouldn't have when resising.");
}

/*	Function Name: Layout
 *	Description: lays out the item in the list.
 *	Arguments: w - the widget.
 *                 xfree, yfree - TRUE if we are free to resize the widget in
 *                                this direction.
 *                 width, height - the is the current width and height that 
 *                                 we are going to layout the threeList widget to,
 *                                 depending on xfree and yfree of course.
 *                               
 *	Returns: TRUE if width or height have been changed.
 */

static Boolean
Layout(w, xfree, yfree, width, height)
Widget w;
Boolean xfree, yfree;
Dimension *width, *height;
{
    ThreeListWidget lw = (ThreeListWidget) w;
    Boolean change = FALSE;
    
/* 
 * If force columns is set then always use number of columns specified
 * by default_cols.
 */

    if (lw->threeList.force_cols) {
        lw->threeList.ncols = lw->threeList.default_cols;
	if (lw->threeList.ncols <= 0) lw->threeList.ncols = 1;
	/* 12/3 = 4 and 10/3 = 4, but 9/3 = 3 */
	lw->threeList.nrows = ( ( lw->threeList.nitems - 1) / lw->threeList.ncols) + 1 ;
	if (xfree) {		/* If allowed resize width. */
	    *width = lw->threeList.ncols * lw->threeList.col_width 
	           + 2 * lw->threeList.internal_width;
	    change = TRUE;
	}
	if (yfree) {		/* If allowed resize height. */
	    *height = (lw->threeList.nrows * lw->threeList.row_height)
                    + 2 * lw->threeList.internal_height;
	    change = TRUE;
	}
	return(change);
    }

/*
 * If both width and height are free to change the use default_cols
 * to determine the number columns and set new width and height to
 * just fit the window.
 */

    if (xfree && yfree) {
        lw->threeList.ncols = lw->threeList.default_cols;
	if (lw->threeList.ncols <= 0) lw->threeList.ncols = 1;
	lw->threeList.nrows = ( ( lw->threeList.nitems - 1) / lw->threeList.ncols) + 1 ;
        *width = lw->threeList.ncols * lw->threeList.col_width
	       + 2 * lw->threeList.internal_width;
	*height = (lw->threeList.nrows * lw->threeList.row_height)
                + 2 * lw->threeList.internal_height;
	change = TRUE;
    }
/* 
 * If the width is fixed then use it to determine the number of columns.
 * If the height is free to move (width still fixed) then resize the height
 * of the widget to fit the current list exactly.
 */
    else if (!xfree) {
        lw->threeList.ncols = ( (int)(*width - 2 * lw->threeList.internal_width)
	                    / (int)lw->threeList.col_width);
	if (lw->threeList.ncols <= 0) lw->threeList.ncols = 1;
	lw->threeList.nrows = ( ( lw->threeList.nitems - 1) / lw->threeList.ncols) + 1 ;
	if ( yfree ) {
  	    *height = (lw->threeList.nrows * lw->threeList.row_height)
		    + 2 * lw->threeList.internal_height;
	    change = TRUE;
	}
    }
/* 
 * The last case is xfree and !yfree we use the height to determine
 * the number of rows and then set the width to just fit the resulting
 * number of columns.
 */
    else if (!yfree) {		/* xfree must be TRUE. */
        lw->threeList.nrows = (int)(*height - 2 * lw->threeList.internal_height) 
	                 / (int)lw->threeList.row_height;
	if (lw->threeList.nrows <= 0) lw->threeList.nrows = 1;
	lw->threeList.ncols = (( lw->threeList.nitems - 1 ) / lw->threeList.nrows) + 1;
	*width = lw->threeList.ncols * lw->threeList.col_width 
	       + 2 * lw->threeList.internal_width;
	change = TRUE;
    }      
    return(change);
}

/*	Function Name: Notify
 *	Description: Notifies the user that a button has been pressed, and
 *                   calles the callback, if the XtNpasteBuffer resource
 *                   is true then the name of the item is also put in the
 *                   X cut buffer ( buf (0) ).
 *	Arguments: w - the widget that the notify occured in.
 *                 event - event that caused this notification.
 *                 params, num_params - not used.
 *	Returns: none.
 */

/* ARGSUSED */
static void
Notify(w, event, params, num_params)
Widget w;
XEvent * event;
String * params;
Cardinal *num_params;
{
    ThreeListWidget lw = ( ThreeListWidget ) w;
    int item, item_len;
    XawThreeListReturnStruct ret_value;
/* 
 * Find item and if out of range then unhighlight and return. 
 * 
 * If the current item is unhighlighted then the user has aborted the
 * notify, so unhighlight and return.
 */

    if ( ((CvtToItem(w, event->xbutton.x, event->xbutton.y, &item))
	  == OUT_OF_RANGE) || (lw->threeList.highlight != item) ) {
        XawThreeListUnhighlight(w);
        return;
    }

    item_len = strlen(lw->threeList.list[item]);

    if ( lw->threeList.paste )	/* if XtNpasteBuffer set then paste it. */
        XStoreBytes(XtDisplay(w), lw->threeList.list[item], item_len);

/* 
 * Call Callback function.
 */

    ret_value.string = lw->threeList.list[item];
    ret_value.list_index = item;
    ret_value.event = event->xbutton.state;
    
    XtCallCallbacks( w, XtNcallback, (XtPointer) &ret_value);
}

/*	Function Name: Unset
 *	Description: unhighlights the current element.
 *	Arguments: w - the widget that the event occured in.
 *                 event - not used.
 *                 params, num_params - not used.
 *	Returns: none.
 */

/* ARGSUSED */
static void
Unset(w, event, params, num_params)
Widget w;
XEvent * event;
String * params;
Cardinal *num_params;
{
  XawThreeListUnhighlight(w);
}

/*	Function Name: Set
 *	Description: Highlights the current element.
 *	Arguments: w - the widget that the event occured in.
 *                 event - event that caused this notification.
 *                 params, num_params - not used.
 *	Returns: none.
 */

/* ARGSUSED */
static void
Set(w, event, params, num_params)
Widget w;
XEvent * event;
String * params;
Cardinal *num_params;
{
  int item;
  ThreeListWidget lw = (ThreeListWidget) w;

  if ( (CvtToItem(w, event->xbutton.x, event->xbutton.y, &item))
      == OUT_OF_RANGE)
    XawThreeListUnhighlight(w);		        /* Unhighlight current item. */
  else if ( lw->threeList.is_highlighted != item )   /* If this item is not */
    XawThreeListHighlight(w, item);	                /* highlighted then do it. */
}

/*
 * Set specified arguments into widget
 */

static Boolean 
SetValues(current, request, new)
Widget current, request, new;
{
    ThreeListWidget cl = (ThreeListWidget) current;
    ThreeListWidget rl = (ThreeListWidget) request;
    ThreeListWidget nl = (ThreeListWidget) new;
    Boolean redraw = FALSE;

    if ((cl->threeList.foreground != rl->threeList.foreground) ||
	(cl->core.background_pixel != rl->core.background_pixel) ||
	(cl->threeList.font != rl->threeList.font) ) {
	XGCValues values;
	XGetGCValues(XtDisplay(current), cl->threeList.graygc, GCTile, &values);
	XmuReleaseStippledPixmap(XtScreen(current), values.tile);
	XtReleaseGC(current, cl->threeList.graygc);
	XtReleaseGC(current, cl->threeList.revgc);
	XtReleaseGC(current, cl->threeList.normgc);
        GetGCs(new);
        redraw = TRUE;
    }

    /* Reset row height. */

    if ((cl->threeList.row_space != rl->threeList.row_space) ||
	(cl->threeList.font != rl->threeList.font)) 
        nl->threeList.row_height = nl->threeList.font->max_bounds.ascent
	                    + nl->threeList.font->max_bounds.descent
			    + nl->threeList.row_space;
    
    if ((cl->core.width != rl->core.width)                     ||
	(cl->core.height != rl->core.height)                   ||
	(cl->threeList.internal_width != rl->threeList.internal_width)   ||
	(cl->threeList.internal_height != rl->threeList.internal_height) ||
	(cl->threeList.column_space != rl->threeList.column_space)       ||
	(cl->threeList.row_space != rl->threeList.row_space)             ||
	(cl->threeList.default_cols != rl->threeList.default_cols)       ||
	(  (cl->threeList.force_cols != rl->threeList.force_cols) &&
	   (rl->threeList.force_cols != rl->threeList.ncols) )           ||
	(cl->threeList.vertical_cols != rl->threeList.vertical_cols)     ||
	(cl->threeList.longest != rl->threeList.longest)                 ||
	(cl->threeList.nitems != rl->threeList.nitems)                   ||
	(cl->threeList.font != rl->threeList.font)                       ||
	(cl->threeList.list != rl->threeList.list)                        ) {

      ResetThreeList(new, TRUE, TRUE);
      redraw = TRUE;
    }

    if (cl->threeList.list != rl->threeList.list)
	nl->threeList.is_highlighted = nl->threeList.highlight = NO_HIGHLIGHT;

    if ((cl->core.sensitive != rl->core.sensitive) ||
	(cl->core.ancestor_sensitive != rl->core.ancestor_sensitive)) {
        nl->threeList.highlight = NO_HIGHLIGHT;
	redraw = TRUE;
    }
    
    if (!XtIsRealized(current))
      return(FALSE);
      
    return(redraw);
}

static void Destroy(w)
    Widget w;
{
    ThreeListWidget lw = (ThreeListWidget) w;
    XGCValues values;
    
    XGetGCValues(XtDisplay(w), lw->threeList.graygc, GCTile, &values);
    XmuReleaseStippledPixmap(XtScreen(w), values.tile);
    XtReleaseGC(w, lw->threeList.graygc);
    XtReleaseGC(w, lw->threeList.revgc);
    XtReleaseGC(w, lw->threeList.normgc);
}

/* Exported Functions */

/*	Function Name: XawThreeListChange.
 *	Description: Changes the list being used and shown.
 *	Arguments: w - the threeList widget.
 *                 list - the new list.
 *                 nitems - the number of items in the list.
 *                 longest - the length (in Pixels) of the longest element
 *                           in the list.
 *                 resize - if TRUE the the threeList widget will
 *                          try to resize itself.
 *	Returns: none.
 *      NOTE:      If nitems of longest are <= 0 then they will be calculated.
 *                 If nitems is <= 0 then the list needs to be NULL terminated.
 */

void
XawThreeListChange(w, list, nitems, longest, resize_it)
Widget w;
char ** list;
int nitems, longest;
Boolean resize_it;
{
    ThreeListWidget lw = (ThreeListWidget) w;

    lw->threeList.list = list;

    if (nitems <= 0) nitems = 0;
    lw->threeList.nitems = nitems;
    if (longest <= 0) longest = 0;
    lw->threeList.longest = longest;

    ResetThreeList(w, resize_it, resize_it);
    lw->threeList.is_highlighted = lw->threeList.highlight = NO_HIGHLIGHT;
    if ( XtIsRealized(w) )
      Redisplay(w, NULL, NULL);
}

/*	Function Name: XawThreeListUnhighlight
 *	Description: unlights the current highlighted element.
 *	Arguments: w - the widget.
 *	Returns: none.
 */

void
XawThreeListUnhighlight(w)
Widget w;
{
    ThreeListWidget lw = ( ThreeListWidget ) w;

    lw->threeList.highlight = NO_HIGHLIGHT;
    if (lw->threeList.is_highlighted != NO_HIGHLIGHT)
        PaintItemName(w, lw->threeList.is_highlighted); /* unhighlight this one. */
}

/*	Function Name: XawThreeListHighlight
 *	Description: Highlights the given item.
 *	Arguments: w - the threeList widget.
 *                 item - the item to hightlight.
 *	Returns: none.
 */

void
XawThreeListHighlight(w, item)
Widget w;
int item;
{
    ThreeListWidget lw = ( ThreeListWidget ) w;
    
    if (XtIsSensitive(w)) {
        lw->threeList.highlight = item;
        if (lw->threeList.is_highlighted != NO_HIGHLIGHT)
            PaintItemName(w, lw->threeList.is_highlighted);  /* Unhighlight. */
	PaintItemName(w, item); /* HIGHLIGHT this one. */ 
    }
}

/*	Function Name: XawThreeListShowCurrent
 *	Description: returns the currently highlighted object.
 *	Arguments: w - the threeList widget.
 *	Returns: the info about the currently highlighted object.
 */

XawThreeListReturnStruct *
XawThreeListShowCurrent(w)
Widget w;
{
    ThreeListWidget lw = ( ThreeListWidget ) w;
    XawThreeListReturnStruct * ret_val;

    ret_val = (XawThreeListReturnStruct *) XtMalloc (sizeof (XawThreeListReturnStruct));
    
    ret_val->list_index = lw->threeList.highlight;
    ret_val->event = 0;
    if (ret_val->list_index == XAW_LIST_NONE)
      ret_val->string = "";
    else
      ret_val->string = lw->threeList.list[ ret_val->list_index ];

    return(ret_val);
}


/*
 * List Widget Creation Routines and stuff.
 */

static void list_callback(w, data, call_data)
Widget w;
XtPointer data;
XtPointer call_data;
{
  ListInfo *li = (ListInfo *)data;
  XawThreeListReturnStruct *list = (XawThreeListReturnStruct *)call_data;

  if (li->func)
    li->func(w, list->string, list->list_index, list->event, li->data);
}



Widget MakeThreeList(item_list, width, height, func, data)
char **item_list;
int width;
int height;
ThreeListCB func;
void *data;
{
  int    n = 0;
  Arg    wargs[10];		/* Used to set widget resources */
  Widget list, vport;
  ListInfo *li;

  if (lsx_curwin->toplevel == NULL && OpenDisplay(0, NULL) == 0)
    return NULL;

  n = 0;
  XtSetArg(wargs[n], XtNwidth,  width);             n++;
  XtSetArg(wargs[n], XtNheight, height);            n++;
  XtSetArg(wargs[n], XtNallowVert, True);           n++;
  XtSetArg(wargs[n], XtNallowHoriz, True);          n++;
  XtSetArg(wargs[n], XtNuseBottom, True);           n++;

  vport = XtCreateManagedWidget("vport", viewportWidgetClass,
			       lsx_curwin->form_widget,wargs,n);
  if (vport == NULL)
    return NULL;

  n = 0;
  XtSetArg(wargs[n], XtNlist,   item_list);         n++;
  XtSetArg(wargs[n], XtNverticalList, True);        n++;
  XtSetArg(wargs[n], XtNforceColumns, True);        n++;
  XtSetArg(wargs[n], XtNdefaultColumns, 1);         n++;
  XtSetArg(wargs[n], XtNborderWidth, 1);            n++;
  
  /*
   * Here we create the list widget and make it the child of the
   * viewport widget so that the viewport will properly handle scrolling
   * it and all that jazz.
   */
  list = XtCreateManagedWidget("threeList", threeListWidgetClass,
			       vport,wargs,n);
  if (list == NULL)
   {
     XtDestroyWidget(vport);
     return NULL;
   }

  li = (ListInfo *)malloc(sizeof(ListInfo));
  if (li == NULL)
   {
     XtDestroyWidget(list);
     XtDestroyWidget(vport);
     return NULL;
   }

  li->func = func;
  li->data = data;
  li->w    = list;

  li->next = scroll_lists;
  scroll_lists = li;

  if (func)
    XtAddCallback(list, XtNcallback, list_callback, li);

  return list;
}    /* end of MakeScrollList() */


